open Atom
open Types
open Coercions

(* There is no polymorphic recursion: recursive functions have no type
   arguments. Ideally, this should be fixed: a recursive function
   should be able to accept an arbitrary number of type arguments. *)

(* Term variables are represented by atoms. Terms are constructed by
   [Internalize], which guarantees that all atoms are unique and that
   there are no dangling free atoms. *)

(* Terms are annotated with source code locations, which are used when
   reporting type errors. *)

type fterm =
  | TeVar of atom
      (* x *)
  | TeRecFun of atom * atom * ftype * ftype * loc_fterm
      (* fun f (x : T) : T = t *)
  | TeFun of atom * ftype * ftype option * loc_fterm
      (* fun (x : T) : T = t *)
  | TeApp of loc_fterm * loc_fterm
      (* t t *)
  | TeLet of atom * loc_fterm * loc_fterm
      (* let x = t in t *)
  | TeTyAbs of atom * loc_fterm
      (* fun a -> t *)
  | TeTyApp of loc_fterm * ftype
      (* t [T] *)
  | TeRecord of loc_fterm AtomMap.t
      (* { L = t; ...; L = t } *)
  | TeProj of loc_fterm * atom
      (* t.L *)
  | TePack of loc_fterm * ftype_context
      (* pack t : exists a. T *)
  | TeUnpack of atom * atom * loc_fterm * loc_fterm
      (* unpack a, x = t in t *)
  | TeCoerce of loc_coercion * loc_fterm
      (* <c> t *)

and loc_fterm =
    Error.location * fterm

(* ------------------------------------------------------------------------- *)

(* Helpers for constructing terms with dummy locations. *)

let dummy =
  Error.dummy

let var x =
  dummy, TeVar x

let recabs f x domain codomain body =
  dummy, TeRecFun (f, x, domain, codomain, body)

let abs x domain  body =
  dummy, TeFun (x, domain, None, body)

let app term1 term2 =
  dummy, TeApp (term1, term2)

let def x term1 term2 =
  dummy, TeLet (x, term1, term2)

let tyabs a body =
  dummy, TeTyAbs (a, body)

let tyapp t ty =
  dummy, TeTyApp (t, ty)

let record fields =
  dummy, TeRecord fields

let proj term label =
  dummy, TeProj (term, label)

let pack term existential =
  dummy, TePack (term, existential)

let unpack a x t1 t2 =
  dummy, TeUnpack (a, x, t1, t2)



(* ------------------------------------------------------------------------- *)

(* Determining a term's free term variables. *)

let rec fv term =
  match term with
  | TeVar x ->
      AtomSet.singleton x
  | TeFun (x, _, _, body) ->
      AtomSet.remove x (loc_fv body)
  | TeRecFun (f, x, _, _, body) ->
      AtomSet.remove f (AtomSet.remove x (loc_fv body))
  | TeApp (term1, term2) ->
      AtomSet.union (loc_fv term1) (loc_fv term2)
  | TeLet (x, term1, term2)
  | TeUnpack (_, x, term1, term2) ->
      AtomSet.union (loc_fv term1) (AtomSet.remove x (loc_fv term2))
  | TeTyAbs (_, term)
  | TeTyApp (term, _)
  | TeProj (term, _)
  | TePack (term, _) ->
      loc_fv term
  | TeRecord fields ->
      AtomMap.fold (fun _ term accu ->
        AtomSet.union (loc_fv term) accu) fields AtomSet.empty
  | TeCoerce (c, term) -> 
      loc_fv term
      

and loc_fv (_, term) =
  fv term

